# rubocop:disable Lint/HandleExceptions
# rubocop:disable Metrics/MethodLength

module Postjob::CLI
  private

  def hosts_query(limit:)
    limit = Integer(limit)

    sql = <<-SQL
      SELECT
        hosts.id,
        hosts.created_at,
        hosts.status,
        heartbeats.attributes AS heartbeat,
        heartbeats.created_at AS heartbeat_created_at,
        hosts.attributes
      FROM postjob.hosts hosts
      LEFT JOIN (
        SELECT
          host_id,
          attributes,
          created_at,
          rank() OVER (PARTITION BY host_id ORDER BY created_at DESC) AS rank
        FROM postjob.events events
        WHERE name='heartbeat'
      ) heartbeats ON heartbeats.host_id=hosts.id AND rank = 1
    SQL

    scope = Simple::SQL::Scope.new(sql)
    scope
      .order_by("hosts.created_at DESC NULLS LAST")
      .paginate(per: limit, page: 1)
  end

  public

  # Show hosts status
  #
  # This command lists all worker_sessions currently in the system.
  #
  # Example:
  #
  #   postjob hosts
  def hosts(limit: "30")
    expect! limit => /\A\d+\z/
    limit = Integer(limit)

    connect_to_database!

    query = hosts_query(limit: limit)

    print_results query: query
  end

  # Show up-to-date hosts information once per second
  def hosts_top(limit: "30")
    loop do
      system "clear"
      hosts(limit: limit)
      sleep 1
    end
  rescue Interrupt
  end

  # resets the host id
  def host_reset
    Postjob::Host.clear_storage
  end

  # Set the host to shutdown state
  def host_shutdown
    connect_to_database!

    Postjob::Host.shutdown!(host_id: ::Postjob.host_id)
  end

  # Set the host to running again
  def host_restart
    connect_to_database!

    Simple::SQL.ask "UPDATE postjob.hosts SET status='running' WHERE id=$1::uuid", ::Postjob.host_id
  end

  def hosts_zombies
    Simple::SQL.print <<~SQL
        SELECT hosts.*
        FROM postjob.hosts hosts
        LEFT JOIN (
          SELECT id, host_id
          FROM postjob.events events
          WHERE name='heartbeat'
            AND created_at > now() at time zone 'utc' - interval '5 minutes'
        ) heartbeats ON hosts.id=heartbeats.host_id
      WHERE status IN ('running', 'shutdown')
        AND heartbeats.id IS NULL
    SQL
  end
end
