# rubocop:disable Lint/HandleExceptions
# rubocop:disable Metrics/MethodLength
# rubocop:disable Metrics/ModuleLength
# rubocop:disable Metrics/PerceivedComplexity
# rubocop:disable Metrics/ParameterLists

module Postjob::CLI
  private

  def ps_query(tags:, limit:, queue:, only_root: false)
    limit = Integer(limit)

    sql = <<-SQL
      SELECT
        postjobs.id,
        full_id,
        postjobs.queue,
        workflow
          || (CASE WHEN workflow_version != '' THEN '@' ELSE '' END)
          || workflow_version
          || (CASE WHEN workflow_method != 'run' THEN '.' || workflow_method ELSE '' END)
          || args AS job,
        CASE
          WHEN (status = 'err') THEN 'err(' || failed_attempts || '/' || max_attempts || ')'
          WHEN (status = 'failed') THEN 'fail(' || failed_attempts || '/' || max_attempts || ')'
          ELSE status::varchar
          END AS status,
        error,
        COALESCE((results->0)::varchar, error_message) AS result,
        cron_interval AS cron,
        iff(is_sticky, COALESCE(substring(sticky_host_id::varchar for 9) || '...', 'yes')::varchar, 'no'::varchar) AS sticky,
        iff(is_greedy, 'yes', 'no'::varchar) AS greedy,
        next_run_at,
        -- to_char(EXTRACT(EPOCH FROM (next_run_at - now() at time zone 'utc')), '999999999.99') AS next_run_in,
        to_char(
          EXTRACT(EPOCH FROM
          (
            COALESCE(
              (SELECT MIN(created_at) FROM postjob.events WHERE postjob_id=postjobs.id AND name IN ('ok', 'failed')),
              now() at time zone 'utc'
            )
            -
            (SELECT MIN(created_at) FROM postjob.events WHERE postjob_id=postjobs.id AND name IN ('processing'))
          )
          ), '999999999.99'
        ) AS processing,

        to_char(EXTRACT(EPOCH FROM (now() at time zone 'utc' - postjobs.created_at)), '999999999.99') AS age,
        tags
      FROM postjob.postjobs AS postjobs
    SQL

    scope = Simple::SQL::Scope.new(sql)
    scope = scope.where("tags @> ?", Postjob::Queue::Encoder.encode(parse_tags(tags))) if tags
    scope = scope.where(queue: queue) if queue
    scope = scope.where("id = root_id") if only_root

    scope
      .paginate(per: limit, page: 1)
      .order_by("root_id DESC, id ASC")
  end

  public

  # Show job status
  #
  # This command lists the statuses of all jobs that are either root jobs,
  # i.e. enqueued workflows, or that have failed.
  #
  # Example:
  #
  #   postjob ps --tags=foo:bar,bar:baz --limit=100
  #
  # For a listing of all jobs in the system use ps:full, see 'postjob help ps:full'
  # for details.
  def ps(*ids, limit: "30", tags: nil, queue: nil, only_root: nil)
    expect! limit => /\A\d+\z/
    limit = Integer(limit)
    queue = queue.split(",") if queue

    connect_to_database!

    # check for timed out and zombie processes
    # ::Postjob::Queue.checkout(nil)

    unless ids.empty?
      ps_full(*ids, limit: limit, tags: tags)
      return
    end

    query = ps_query(tags: tags, limit: limit, queue: queue, only_root: only_root)
    query = query.where("root_id=id OR status NOT IN ('sleep', 'ok') OR failed_attempts > 0")

    print_results query: query
  end

  def ps_full(*ids, limit: "30", tags: nil, queue: nil)
    queue = queue.split(",") if queue

    connect_to_database!

    query = ps_query(tags: tags, limit: limit, queue: queue)

    unless ids.empty?
      parsed_ids = parse_ids(*ids)
      query = query.where("root_id IN (#{parsed_ids.join(',')})")
    end

    print_results query: query, on_empty: "Note that ps_full requires the **root ids**"
  end

  # Show up-to-date information once per second
  def ps_top(*ids, limit: "30", tags: nil, full: false, queue: nil, only_root: false)
    loop do
      system "clear"
      if full
        ps_full(*ids, limit: limit, tags: tags, queue: queue)
      else
        ps(*ids, limit: limit, tags: tags, queue: queue, only_root: only_root)
      end
      sleep 1
    end
  rescue Interrupt
  end

  # Print the result of a job
  def ps_result(id)
    connect_to_database!

    scope = Simple::SQL::Scope.new("SELECT * FROM postjob.postjobs")
    scope = scope.where("postjobs.id = ANY(?)", parse_ids(id))

    job = Simple::SQL.ask(scope, into: Postjob::Job)
    raise "Job ##{id}: no such job" unless job

    if job.status == "timeout" || job.status == "failed"
      # resolve raises exceptions on failing jobs. At this point we'll just tell
      # the user to be aware of that, and let job.resolve fail below.
      logger.warn "Job failed. Details below."
    end

    result = job.resolve

    # Pending jobs don't have a result yet. Just saying..
    raise "Job ##{id}: job is still pending" if result == :pending

    # Print result. If this is a structure we pretty print, if not we
    # print verbatim.
    if result.is_a?(String) || !result.respond_to?(:empty?)
      puts result
    else
      puts JSON.pretty_generate(result)
    end
  end

  # Show all information about this job
  def ps_show(id, *ids)
    connect_to_database!

    scope = Simple::SQL::Scope.new <<~SQL
      SELECT postjobs.*, tokens.token
      FROM postjob.postjobs AS postjobs
      LEFT JOIN postjob.tokens AS tokens ON tokens.postjob_id=postjobs.id
    SQL
    scope = scope.where("postjobs.id = ANY(?)", parse_ids(id, *ids))

    Simple::SQL.all(scope, into: Postjob::Job) do |job|
      pp job
    end
  end
end
