module Postjob::CLI
  # Run a single job
  def step(count: 1, queue: nil, host_id: nil)
    run count: count, queue: queue, host_id: host_id, heartbeat: false
  end

  # Start the control connection
  #
  # The control connection runs the heartbeat, and checks for jobs on the
  # control queue.
  def run_control(host_id: nil)
    # By running the "control" queue this code is quite efficient, it does
    # not poll the database.
    _run(queue: "control", host_id: host_id, heartbeat: true)

    host_id ||= Postjob.host_id

    # Poll until there are no more working sessions.
    # [TODO] - improve by listening.
    wait_for_no_running_session(host_id: host_id)

    run_control_shutdown(host_id: host_id)
  end

  private

  # This is called after the control connection did shut down
  #
  # This method can be reimplemented in a plugin to allow shutting down
  # worker machines.
  def run_control_shutdown(host_id: nil)
    Postjob.logger.success "postjob control:shutdown host_id: #{host_id}"
  end

  public

  # Run postjobs.
  #
  # This method runs jobs as they become ready.
  #
  # Parameters:
  #
  # - --count=<count> maximum number of jobs to process. Default: unlimited.
  # - --queue=queue1,queue2,queue3 run only the specified queues.
  # - --heartbeat=no don't start heartbeat process.
  #
  def run(count: nil, queue: nil, fast: false, host_id: nil)
    count = Integer(count) if count
    processed = _run(count: count, queue: queue, fast: fast, host_id: host_id, heartbeat: false)

    if !count || processed < count
      # The runner has been shut down externally. Wait for interrupt.
      Postjob.logger.success "External shut down initiated."
      STDERR.puts "External shut down initiated. Press ^C to terminate process."
      STDIN.read
    end
  end

  private

  def _run(count: nil, queue:, fast: false, host_id:, heartbeat:)
    expect! Integer(host_id, 16) => 1..0xffffffff if host_id

    expect! heartbeat => [ true, false ]

    Postjob.fast_mode = (fast ? true : false)

    connect_to_database!

    if host_id
      Postjob::Host.host_id = "%08x-0000-0000-0000-000000000000" % Integer(host_id, 16)
      Postjob.logger.info "Using host_id: #{Postjob::Host.host_id}"
    end

    processed = Postjob.run(count: count, queues: queue&.split(","), heartbeat: heartbeat) do |job_id|
      logger.info "Processed job w/id #{job_id}" if job_id
    end

    logger.info "Processed #{processed} jobs"

    processed
  end

  def wait_for_no_running_session(host_id:)
    Postjob.logger.debug "Waiting for shutdown"
    loop do
      count = Simple::SQL.ask "SELECT COUNT (*) FROM postjob.worker_sessions WHERE id=$1", host_id
      Postjob.logger.info "#{host_id}: #{count} running sessions"

      break if count == 0
      sleep 0.2
    end

    Postjob.logger.info "#{host_id}: no more running sessions"
  end
end
