# rubocop:disable Lint/HandleExceptions
# rubocop:disable Metrics/MethodLength

module Postjob::CLI
  private

  def sessions_query(limit:)
    limit = Integer(limit)

    sql = <<-SQL
      SELECT
        worker_sessions.id,
        (substring(worker_sessions.host_id::varchar for 9) || '...') AS host_id,
        worker_sessions.status,
        array_to_string(worker_sessions.queues, ', ') AS queues,
        worker_sessions.client_socket,
        worker_sessions.workflows,
        worker_sessions.created_at,
        job_event.name AS event_name,
        job_event.created_at AS event_created_at,
        heartbeat.attributes AS heartbeat,
        heartbeat.created_at AS heartbeat_created_at
      FROM postjob.worker_sessions AS worker_sessions
      LEFT JOIN (
        SELECT
          worker_sessions.id,
          MAX(events.id) AS event_id
        FROM postjob.worker_sessions
        LEFT JOIN postjob.events events ON events.worker_session_id=worker_sessions.id
        WHERE events.name != 'heartbeat'
        GROUP BY worker_sessions.id
      ) last_job_event ON last_job_event.id=worker_sessions.id
      LEFT JOIN postjob.events job_event ON job_event.id=last_job_event.event_id
      LEFT JOIN (
        SELECT
          worker_sessions.id,
          MAX(events.id) AS event_id
        FROM postjob.worker_sessions
        LEFT JOIN postjob.events events ON events.worker_session_id=worker_sessions.id
        WHERE events.name = 'heartbeat'
        GROUP BY worker_sessions.id
      ) last_heartbeat ON last_heartbeat.id=worker_sessions.id
      LEFT JOIN postjob.events heartbeat ON heartbeat.id=last_heartbeat.event_id
    SQL

    scope = Simple::SQL::Scope.new(sql)
    scope = scope.where("worker_sessions.id != '00000000-0000-0000-0000-000000000000'::uuid")
    scope
      .paginate(per: limit, page: 1)
      .order_by("heartbeat_created_at DESC NULLS LAST")
  end

  public

  # Show sessions status
  #
  # This command lists all worker sessions currently in the system.
  #
  # Example:
  #
  #   postjob sessions
  def sessions(limit: "30")
    expect! limit => /\A\d+\z/
    limit = Integer(limit)

    connect_to_database!

    # check for timed out and zombie processes
    # ::Postjob::Queue.checkout(nil)

    query = sessions_query(limit: limit)

    print_results query: query
  end

  # Show up-to-date session information once per second
  def sessions_top(limit: "30")
    loop do
      system "clear"
      sessions(limit: limit)
      sleep 1
    end
  rescue Interrupt
  end
end
