-- hosts ----------------------------------------------------------------------

DO $$
  BEGIN
    CREATE TYPE {SCHEMA_NAME}.host_statuses AS ENUM (
      'running',   -- host is running
      'shutdown',  -- host is shutting down
      'stopped'    -- host is not running
    );
  EXCEPTION
    WHEN duplicate_object THEN RAISE DEBUG 'type {SCHEMA_NAME}.host_statuses already exists';
  END;
$$;


-- The hosts table records available hosts. hosts are the basis for the
-- sticky jobs feature - a sticky job is a job which can only be checked
-- out by a specific host.
--

CREATE TABLE IF NOT EXISTS {SCHEMA_NAME}.hosts (
  id UUID PRIMARY KEY  DEFAULT (gen_random_uuid()),     -- UUID identifying a worker **host**

  -- The workflow status, one of  'running', 'shutdown', 'stopped'
  status {SCHEMA_NAME}.host_statuses NOT NULL DEFAULT 'stopped',

  attributes JSONB NOT NULL DEFAULT '{}'::JSONB,
  created_at timestamp NOT NULL DEFAULT (now() at time zone 'utc')
);

CREATE INDEX IF NOT EXISTS hosts_attributes_idx
  ON {SCHEMA_NAME}.hosts USING GIN (attributes jsonb_path_ops);

DO $$
  BEGIN
    ALTER TABLE {SCHEMA_NAME}.hosts ADD COLUMN status {SCHEMA_NAME}.host_statuses NOT NULL DEFAULT 'stopped';
  EXCEPTION
    WHEN duplicate_column THEN RAISE DEBUG 'column {SCHEMA_NAME}.hosts.host_statuses already exists';
  END;
$$;

-- returns _null_host_id ------------------------------------------------------

DO $$
  DECLARE
    null_uuid UUID := {SCHEMA_NAME}._null_uuid();
  BEGIN
    IF NOT EXISTS (SELECT 1 FROM {SCHEMA_NAME}.hosts WHERE id = null_uuid) THEN
      INSERT INTO {SCHEMA_NAME}.hosts(id, attributes) VALUES(null_uuid, '{}');
    END IF;
  END;
$$ LANGUAGE plpgsql;

-- host_register: registers a host --------------------------------------------

DROP FUNCTION IF EXISTS  {SCHEMA_NAME}.host_register(p_attrs JSONB);
CREATE OR REPLACE FUNCTION {SCHEMA_NAME}.host_register(p_attrs JSONB, v_id UUID)
  RETURNS UUID
AS $$
  BEGIN
    IF v_id IS NULL THEN
      v_id := gen_random_uuid();
    END IF;

    INSERT INTO {SCHEMA_NAME}.hosts (id, attributes, status)
      VALUES (v_id, p_attrs, 'running')
    ON CONFLICT(id) DO UPDATE SET attributes=p_attrs;
    RETURN v_id;
  END;
$$ LANGUAGE plpgsql;


-- wakeup runners after changing hosts

-- when a host changes its status to shutdown, all of its runners should
-- shutdown quickly.

BEGIN;
  DROP TRIGGER IF EXISTS _wakeup_runners ON {SCHEMA_NAME}.hosts;

  CREATE TRIGGER _wakeup_runners AFTER UPDATE
    ON {SCHEMA_NAME}.hosts
    FOR EACH STATEMENT
    EXECUTE PROCEDURE {SCHEMA_NAME}._wakeup_runners();
COMMIT;
