-- worker_sessions ------------------------------------------------------------

-- The worker_sessions table records available "worker worker_sessions". The
-- following information is generated/recorded:
--
-- - id:            a UUID, which is unique per worker_session.
-- - host_id:       a UUID, which is unique across all workers on a machine.
-- - client_socket: contains "host:port" of a connection
-- - workflows:     an array of workflows with and without version numbers.
--                  This describes which workflows can be checked out in this
--                  session.
--
-- The host_id value will be used for sticky workflows: A workflow with
-- `sticky: true` will only be checked out to workers with the same host_id.
-- This is useful for workflows that use temporary files to share state
-- between jobs.
--
-- The +id+ and +host_id+ UUIDs are generated on the queue. The client is
-- supposed to get a worker_session via +worker_session_start(id, host_id, workflows)+
--


DO $$
  BEGIN
    CREATE TYPE {SCHEMA_NAME}.session_statuses AS ENUM (
      'running',   -- session is running
      'stopped'    -- session is stopped
    );
  EXCEPTION
    WHEN duplicate_object THEN RAISE DEBUG 'type {SCHEMA_NAME}.session_statuses already exists';
  END;
$$;


CREATE TABLE IF NOT EXISTS {SCHEMA_NAME}.worker_sessions (
  id UUID PRIMARY KEY  DEFAULT (gen_random_uuid()),                       -- UUID identifying a worker **process**
  host_id UUID NOT NULL REFERENCES {SCHEMA_NAME}.hosts ON DELETE CASCADE, -- UUID identifying a worker **host**
  client_socket VARCHAR,                                                  -- host:port of connection (from pg_stat_activity)
  workflows VARCHAR[] NOT NULL,                                           -- array of workflow versions available on that worker
  queues VARCHAR[] NOT NULL,                                              -- array of queue names available on that worker
  attributes JSONB NOT NULL DEFAULT '{}'::JSONB,
  status {SCHEMA_NAME}.session_statuses NOT NULL DEFAULT 'running',
  created_at timestamp NOT NULL DEFAULT (now() at time zone 'utc')
);

CREATE INDEX IF NOT EXISTS worker_sessions_attributes_idx
  ON {SCHEMA_NAME}.worker_sessions USING GIN (attributes jsonb_path_ops);

-- worker_session_start: starts or reuses a worker_session ----------------------------------

DO $$
  DECLARE
    null_uuid UUID := {SCHEMA_NAME}._null_uuid();
  BEGIN
    IF NOT EXISTS (SELECT 1 FROM {SCHEMA_NAME}.worker_sessions WHERE id = null_uuid) THEN
      INSERT INTO {SCHEMA_NAME}.worker_sessions(id, host_id, workflows, queues) VALUES(null_uuid, null_uuid, '{}', ARRAY[]::varchar[]);
    END IF;
  END;
$$ LANGUAGE plpgsql;

DO $$
  BEGIN
    ALTER TABLE {SCHEMA_NAME}.worker_sessions ADD COLUMN queues VARCHAR[];
    UPDATE {SCHEMA_NAME}.worker_sessions SET queues = Array['{DEFAULT_QUEUE}'] WHERE queues IS NULL;
    ALTER TABLE {SCHEMA_NAME}.worker_sessions ALTER COLUMN queues SET NOT NULL;
  EXCEPTION
    WHEN duplicate_column THEN RAISE DEBUG 'column {SCHEMA_NAME}.worker_sessions.queues already exists';
  END;
$$;


DO $$
  BEGIN
    ALTER TABLE {SCHEMA_NAME}.worker_sessions ADD COLUMN status {SCHEMA_NAME}.session_statuses NOT NULL DEFAULT 'running';
  EXCEPTION
    WHEN duplicate_column THEN RAISE DEBUG 'column {SCHEMA_NAME}.worker_sessions.session_statuses already exists';
  END;
$$;

UPDATE {SCHEMA_NAME}.worker_sessions SET status = 'stopped' WHERE id = {SCHEMA_NAME}._null_uuid();
