CREATE TABLE IF NOT EXISTS {SCHEMA_NAME}.events (
  id          BIGSERIAL PRIMARY KEY,
  postjob_id  BIGINT REFERENCES {SCHEMA_NAME}.postjobs ON DELETE SET NULL,
  host_id     UUID NOT NULL REFERENCES {SCHEMA_NAME}.hosts ON DELETE CASCADE,
  worker_session_id  UUID REFERENCES {SCHEMA_NAME}.worker_sessions ON DELETE SET NULL,
  name        VARCHAR NOT NULL,
  attributes  JSONB not null DEFAULT '{}'::JSONB,
  created_at  timestamp NOT NULL DEFAULT (now() at time zone 'utc')
);

-- We do not create an index on events.attributes just yet. events has a high
-- write rate, and until we know for sure that we'll need this index we
-- better not have it.
--
-- CREATE INDEX IF NOT EXISTS events_attributes_idx
--  ON {SCHEMA_NAME}.events USING GIN (attributes jsonb_path_ops);

CREATE INDEX IF NOT EXISTS events_postjob_id_idx
  ON {SCHEMA_NAME}.events(postjob_id);

--- define triggers to automatically create events -----------------------------

--
-- An event is created whenever a job's status changes.
--

CREATE OR REPLACE FUNCTION {SCHEMA_NAME}._create_postjob_event() RETURNS TRIGGER AS $$
  BEGIN
    IF TG_OP = 'UPDATE' AND (OLD.status = NEW.status) THEN
      RETURN NEW;
    END IF;

    INSERT INTO {SCHEMA_NAME}.events (postjob_id, name, host_id, worker_session_id)
      VALUES (
        NEW.id,
        NEW.status,
        (SELECT host_id FROM {SCHEMA_NAME}.worker_sessions WHERE id=NEW.last_worker_session_id),
        NEW.last_worker_session_id
      );
    RETURN NEW;
  END;
$$ LANGUAGE plpgsql;

BEGIN;
  DROP TRIGGER IF EXISTS _create_postjob_event ON {SCHEMA_NAME}.postjobs;

  CREATE TRIGGER _create_postjob_event AFTER INSERT OR UPDATE
    ON {SCHEMA_NAME}.postjobs
    FOR EACH ROW
    EXECUTE PROCEDURE {SCHEMA_NAME}._create_postjob_event();
COMMIT;

--
-- An event is also created whenever a worker sends in a heartbeat, see
-- 013_worker_sessions_functions.sql for details.
--

--- create an initial set of events -------------------------------------------
--
-- This is a best-effort procedure; we cannot generate data that isn't here.
--

DO $$
DECLARE
  null_uuid UUID := {SCHEMA_NAME}._null_uuid();
BEGIN
  IF NOT EXISTS (SELECT id FROM {SCHEMA_NAME}.events LIMIT 1) THEN
    INSERT INTO {SCHEMA_NAME}.events (postjob_id, name, created_at, host_id, worker_session_id)
    SELECT * FROM (
      SELECT id, 'ready',         created_at AS created_at, null_uuid, null_uuid FROM {SCHEMA_NAME}.postjobs
        UNION
      SELECT id, status::varchar, updated_at AS created_at, null_uuid, null_uuid FROM {SCHEMA_NAME}.postjobs WHERE status != 'ready'
    ) sq ORDER BY created_at ;
  END IF;
END
$$;
