-- zombie checks --------------------------------------------------------------


CREATE OR REPLACE FUNCTION {SCHEMA_NAME}._set_job_zombie(
  job_id BIGINT,
  p_fast_mode BOOLEAN) RETURNS VOID AS $$
DECLARE
  p_worker_session_id UUID;
  session {SCHEMA_NAME}.worker_sessions;
BEGIN
  SELECT * INTO session
    FROM {SCHEMA_NAME}.worker_sessions worker_sessions
    INNER JOIN {SCHEMA_NAME}.postjobs postjobs ON postjobs.last_worker_session_id=worker_sessions.id
    WHERE postjobs.id=job_id;

  p_worker_session_id := {SCHEMA_NAME}._null_uuid();
  PERFORM {SCHEMA_NAME}._reset_job_processing(p_worker_session_id, job_id);

  -- write error info
  UPDATE {SCHEMA_NAME}.postjobs
  SET
    error='Zombie',
    error_message='host ' || session.host_id || ' disappeared',
    error_backtrace=NULL,
    failed_attempts=failed_attempts+1,
    next_run_at=NULL
  WHERE id=job_id;

  -- prepare next run, if any
  PERFORM {SCHEMA_NAME}._prepare_rerun(job_id, 'err', p_fast_mode);
  PERFORM {SCHEMA_NAME}._wakeup_parent_job(p_worker_session_id, job_id);
END;
$$ LANGUAGE plpgsql;

-- This method runs a zombie check. Its result is written into the database as
-- a 'zombie' event. A zombie check should only happen once per minute.
--
-- A zombie is a job whichs runs on a host which hasn't checked in for a while.
-- (This is currently at 5 minutes)

CREATE OR REPLACE FUNCTION {SCHEMA_NAME}._zombie_check(p_fast_mode BOOLEAN) RETURNS int AS $$
DECLARE
  zombie_id bigint;
  _one int;
  zombie_count int;
  zombie_threshold interval := interval '5 minutes';
BEGIN
  zombie_count := 0;
  FOR zombie_id, _one IN
    -- select jobs that have a last_worker_session_id, which points to a host
    -- whose latest heartbeat is older than +zombie_threshold+.
    --
    -- We ignore hosts (and, for that matter, jobs) that don't  have any heartbeats,
    -- since this scenario should only appear during tests.
    SELECT jobs.id, 1
    FROM {SCHEMA_NAME}.postjobs jobs
    LEFT JOIN {SCHEMA_NAME}.worker_sessions sessions ON jobs.last_worker_session_id=sessions.id
    LEFT JOIN
    (
      SELECT host_id, MAX(created_at) AS created_at
      FROM {SCHEMA_NAME}.events
      WHERE name = 'heartbeat'
      GROUP BY host_id
    ) latest_heartbeat ON sessions.host_id = latest_heartbeat.host_id
    WHERE jobs.status IN ('processing')
      AND latest_heartbeat.created_at < ((now() at time zone 'utc') - zombie_threshold)
  LOOP
    PERFORM {SCHEMA_NAME}._set_job_zombie(zombie_id, p_fast_mode);
    zombie_count := zombie_count + 1;
  END LOOP;

  RETURN zombie_count;
END;
$$ LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION {SCHEMA_NAME}.zombie_check(p_fast_mode BOOLEAN)
RETURNS VOID AS $$
DECLARE
  zombie_check_interval interval := '1 minute';
  p_zombie_count int;
BEGIN
  -- In order to properly throttle the zombie check we keep the latest zombie
  -- check result in the database. If this is younger than 1 minute we do not
  -- run the zombie check.
  IF EXISTS (
    SELECT 1 FROM {SCHEMA_NAME}.events
    WHERE name='zombie'
      AND created_at > (now() at time zone 'utc') - zombie_check_interval
  ) THEN
    RETURN;
  END IF;

  -- Historically we do not need zombie_check results with 0 zombies. We do
  -- write them, initially, for the throttling to be effective, but later on
  -- we can remove them again. We only look for events younger then a 1 hour
  -- threshold, for optimization reasons.

  DELETE FROM {SCHEMA_NAME}.events
    WHERE name='zombie'
      AND created_at > (now() at time zone 'utc') - interval '1 hour'
      AND attributes @> '{"zombie_count": 0}';
  
  -- run the zombie checker and keep the result.
  p_zombie_count := {SCHEMA_NAME}._zombie_check(p_fast_mode);

  INSERT INTO {SCHEMA_NAME}.events(name, host_id, attributes)
    VALUES('zombie', {SCHEMA_NAME}._null_uuid(), jsonb_build_object('zombie_count', p_zombie_count));
END;
$$ LANGUAGE plpgsql;
