require_relative "./database_info"

module Postjob::Queue::Search
  class ScopeBuilder
    DatabaseInfo = ::Postjob::Queue::DatabaseInfo

    SCHEMA_NAME = ::Postjob::Queue::SCHEMA_NAME

    class << self
      def build(model:, filter:, attributes:)
        attributes ||= default_search_attributes(model)

        builder(model).send(:scope, filter: filter, attributes: attributes)
      end

      private

      # describes which attributes should be returned in addition to a model's columns.
      EXTRA_SEARCH_ATTRIBUTES = {
        "postjobs" => [
          "workflow || COALESCE('@' || workflow_version, '') || args AS job",
          "COALESCE((results->0)::varchar, error_message) AS result",
          "(now() at time zone 'utc') - created_at AS age",
          "updated_at - created_at AS runtime"
        ]
      }

      def default_search_attributes(model)
        DatabaseInfo.table_columns("#{SCHEMA_NAME}.#{model}") + (EXTRA_SEARCH_ATTRIBUTES[model] || [])
      end

      # Returns the name of the column used for dynamic filtering.
      def dynamic_filter(model)
        case model
        when "postjobs" then "tags"
        else "attributes"
        end
      end

      def builder(model)
        @builders ||= {}
        @builders[model] ||= new(model: model, dynamic_filter: dynamic_filter(model))
      end
    end

    private

    def initialize(model:, dynamic_filter:)
      @model      = model
      @table_name = "#{SCHEMA_NAME}.#{model}"
      @column_types = DatabaseInfo.column_types(@table_name)
      @dynamic_filter = dynamic_filter
    end

    def scope(filter:, attributes:)
      expect! filter => [Hash, nil]
      expect! attributes => Array

      # build Scope
      scope = Simple::SQL::Scope.new("SELECT #{attributes.join(", ")} FROM #{@table_name}")
      scope = apply_filters(scope, filter) if filter && !filter.empty?
      scope
    end

    # apply the filter hash onto the passed in scope. This matches all filters
    # with a name which is matching a column name against the column values.
    # It matches every other filter value against an entry in the
    # dynamic_filter column.
    def apply_filters(scope, filter)
      filter.each_key do |key|
        expect! key => [Symbol, String]
      end

      converted_filters = filter.inject({}) do |hsh, (key, value)|
        matches = convert_filter_value(value, key: key)
        matches = matches.first if matches.length == 1
        hsh.update key => matches
      end

      column_filters, tags_filters = converted_filters.partition { |key, _| !@column_types[key].nil? }

      scope = scope.where(Hash[column_filters]) unless column_filters.empty?
      scope = scope.where(@dynamic_filter => Hash[tags_filters]) unless tags_filters.empty?

      scope
    end

    def convert_filter_value(value, key:)
      value = Array(value)

      case @column_types[key]
      when "bigint"   then value.map { |v| Integer(v) }
      when "integer"  then value.map { |v| Integer(v) }
      when nil        then value.map(&:to_s) + value.grep(/\A-?\d+\z/).map(&:to_i)
      else                 value.map(&:to_s)
      end
    end
  end
end
