# The registry holds a list of all available workflows
class Postjob::Registry
  def self.instance
    @instance ||= new
  end

  # Used for tests
  def self.reset! # :nodoc:
    @instance = nil
  end

  def self.workflows
    instance.workflows
  end

  # returns an array with the name of all queues that are configured here.
  def self.queues
    queues = workflows
             .select { |_, spec| spec.runnable? }
             .map(&:last)
             .map(&:options)
             .map(&:queue)

    # For a while "ruby" was the name of the default queue. Since we might
    # have jobs with the queue in the database we always return "ruby" as
    # one of the runnable queues, even if there is no workflow explicitely
    # registered on a "ruby" queue.
    queues << "ruby"
    queues.uniq
  end

  def self.workflow_names
    instance.workflows.keys.map(&:first).uniq
  end

  def self.runnable_workflows_with_versions
    workflows
      .select { |_, spec| spec.runnable? }
      .inject([]) do |ary, (name_and_version, _spec)|
        name, version = *name_and_version
        ary << name << "#{name}#{version}"
      end
  end

  def self.workflows_with_versions
    instance.workflows_with_versions
  end

  def self.register(workflow, options = {})
    instance.register(workflow, options)
  end

  def self.load(glob_pattern)
    Dir.glob(glob_pattern).sort.each do |path|
      before = Postjob::Registry.workflow_names
      Kernel.load path
      after = Postjob::Registry.workflow_names
      new_workflows = after - before
      next if new_workflows.empty?

      Postjob.logger.debug "#{path}: registered workflow(s) #{new_workflows.join(', ')}"
    end
  end

  class WorkflowSpec
    class Options
      DEFAULTS = {
        version: "0.0",
        max_attempts: 5,
        timeout: 15 * 60,
        sticky: false,
        greedy: false,
        cron_interval: nil,
        queue: Postjob::DEFAULT_QUEUE
      }

      attr_reader :version
      attr_reader :max_attempts
      attr_reader :timeout
      attr_reader :sticky
      attr_reader :greedy
      attr_reader :cron_interval
      attr_reader :queue

      def initialize(options)
        expect! options => {
          version:        [ nil, /^\d+(\.\d+)*/ ],
          max_attempts:   [ nil, Integer ],
          timeout:        [ nil, Integer ],
          sticky:         [ nil, true, false ],
          cron_interval:  [ nil, Integer ],
          queue:          [ nil, String ]
        }

        options = DEFAULTS.merge(options)

        options[:sticky] ||= options[:greedy]

        @version        = options[:version]
        @max_attempts   = options[:max_attempts]
        @timeout        = options[:timeout]
        @sticky         = options[:sticky]
        @greedy         = options[:greedy]
        @cron_interval  = options[:cron_interval]
        @queue          = options[:queue]
      end

      def inspect
        r = {}
        r[:version]         = @version
        r[:max_attempts]    = @max_attempts
        r[:timeout]         = @timeout
        r[:sticky]          = @sticky
        r[:greedy]          = @greedy
        r[:cron_interval]   = @cron_interval
        r[:queue]           = @queue

        m = []
        r.each do |k, v|
          next if DEFAULTS[k] == v
          m << "#{k}: #{v.inspect}"
        end
        m.join(", ")
      end
    end

    attr_reader :workflow
    attr_reader :options

    def runnable?
      workflow.is_a?(Module)
    end

    def name
      case workflow
      when Module then workflow.name
      when String then workflow
      end
    end

    def initialize(workflow, options)
      expect! workflow => [ Module, String ]
      @workflow = workflow
      @options  = Options.new(options)
    end

    def supports_cleanup?
      return false unless lookup_workflow_method(:run)
      return false unless lookup_workflow_method(:cleanup)

      true
    end

    private

    def lookup_workflow_method(name)
      return unless workflow.is_a?(Module)
      workflow.method(name)
    rescue NameError
      nil
    end
  end

  # looks up a specific version of a specific workflow. Returns a WorkflowSpec
  # object.
  def self.lookup!(name:, version:)
    instance.lookup! name: name, version: version
  end

  def self.[](name)
    lookup! name: name, version: ""
  rescue KeyError
    nil
  end

  attr_reader :workflows
  attr_reader :workflows_with_versions

  def initialize
    @workflows = {}
    @workflows_with_versions = []
  end

  def register(workflow, options)
    if options[:greedy] && options[:sticky] == false
      raise ArgumentError, "#{workflow}: a greedy job must also be sticky"
    end

    spec = WorkflowSpec.new(workflow, options)

    @workflows_with_versions << spec.name << "#{spec.name}#{spec.options.version}"

    @workflows[[spec.name, ""]] = spec
    @workflows[[spec.name, spec.options.version]] = spec
  end

  # looks up a specific version of a specific workflow. Returns the workflow
  # module itself
  def lookup!(name:, version:)
    expect! name => String
    expect! version => String

    @workflows.fetch([name, version])
  end
end
