# rubocop:disable Metrics/MethodLength

require_relative "./history"

# It seems that vmstat relies on loading specific subclasses automatically and/or deferredly.
# This should be fine, but it results in failing tests sometimes. By explicitely loading
# these we can circumvent this.
require "vmstat"
require "vmstat/disk"
require "vmstat/memory"
require "vmstat/cpu"
require "vmstat/load_average"
require "vmstat/cpu"
require "vmstat/disk"
require "vmstat/linux_disk"
require "vmstat/linux_memory"
require "vmstat/load_average"
require "vmstat/memory"
require "vmstat/network_interface"
require "vmstat/task"

#
# A heartbeat monitor
#
# The heartbeat monitor watches various machine metrics (disk usage, CPU load, network traffic)
# and yields current measurements once a second.
#
#
module Heartbeat # :nodoc:
  # The heartbeat monitor watches various machine metrics (disk usage, CPU load, network traffic)
  # and yields current measurements once every cycle_length_seconds seconds.

  # This method yields hashes with these keys:
  #
  # {
  #   net_errors_1min:    ...   # network errors in the last minute
  #   net_errors_5min:    ...   # network errors in the last 5 minutes
  #   net_errors_15min:   ...   # network errors in the last 15 minutes
  #   net_in_1min:        ...   # incoming network traffic in the last minute (bytes)
  #   net_in_5min:        ...   # incoming network traffic in the last 5 minutes (bytes)
  #   net_in_15min:       ...   # incoming network traffic in the last 15 minutes (bytes)
  #   net_out_1min:       ...   # outgoing network traffic in the last minute (bytes)
  #   net_out_5min:       ...   # outgoing network traffic in the last 5 minutes (bytes)
  #   net_out_15min:      ...   # outgoing network traffic in the last 15 minutes (bytes)
  #
  #   uptime:             ...   # uptime (seconds)
  #   cpu_load_1min:      ...   # cpu load
  #   cpu_load_5min:      ...   # cpu load
  #   cpu_load_15min:     ...   # cpu load
  #
  #   disk_used:          ...   # used disk space, over all disks
  #   disk_available:     ...   # available disk space, over all disks
  # }

  def self.monitor(cycle_length_seconds)
    monitor = Monitor.new(cycle_length_seconds)

    loop do
      measurement = monitor.measure
      do_continue = yield measurement
      break if do_continue == false
      monitor.sleep
    end
  end

  class Monitor # :nodoc:
    def initialize(cycle_length_seconds)
      @history = History.new(15 * 60 / cycle_length_seconds + 1)
      @cycle_length_seconds = cycle_length_seconds

      @count = 0
      @started_at = Time.now
    end

    def measure
      @count += 1
      take_measurement
    end

    def sleep
      sleep_time = (@started_at + @count * @cycle_length_seconds) - Time.now
      Kernel.sleep sleep_time if sleep_time > 0
    end

    private

    def take_measurement
      now = snapshot

      @history << now

      past_1min  = @history.last_nth(60 / @cycle_length_seconds)
      past_5min  = @history.last_nth(5 * 60 / @cycle_length_seconds)
      past_15min = @history.last_nth(15 * 60 / @cycle_length_seconds)

      {
        net_errors_1min:    difference(:network_errors, past_1min, now),
        net_errors_5min:    difference(:network_errors, past_5min, now),
        net_errors_15min:   difference(:network_errors, past_15min, now),

        net_in_1min:        difference(:network_in, past_1min, now),
        net_in_5min:        difference(:network_in, past_5min, now),
        net_in_15min:       difference(:network_in, past_15min, now),

        net_out_1min:       difference(:network_out, past_1min, now),
        net_out_5min:       difference(:network_out, past_5min, now),
        net_out_15min:      difference(:network_out, past_15min, now),

        uptime:             now[:uptime],
        cpu_load_1min:      now[:cpu_load_1min],
        cpu_load_5min:      now[:cpu_load_5min],
        cpu_load_15min:     now[:cpu_load_15min],
        disk_used:          now[:disk_used],
        disk_available:     now[:disk_available]
      }
    end

    def difference(key, past, now)
      return nil unless past
      now[key] + past[key]
    end

    # collect current stats, and aggregate network and disk values into something
    # meaningful
    def snapshot
      snapshot = Vmstat.snapshot

      # [TODO] It would be great if we could filter based on the interface.type value.
      # vmstat's rdoc says "BSD numbers", but I couldn't find documentation.
      relevant_network_interfaces = snapshot.network_interfaces.reject { |interface| interface.name =~ /^lo/ }

      {
        network_in:         sum(relevant_network_interfaces, &:in_bytes),
        network_out:        sum(relevant_network_interfaces, &:out_bytes),
        network_errors:     (sum(relevant_network_interfaces, &:in_errors) + sum(relevant_network_interfaces, &:out_errors)),

        uptime:             snapshot.at - snapshot.boot_time,
        cpu_load_1min:      snapshot.load_average.one_minute,
        cpu_load_5min:      snapshot.load_average.five_minutes,
        cpu_load_15min:     snapshot.load_average.fifteen_minutes,

        disk_used:         sum(snapshot.disks, &:used_bytes),
        disk_available:    sum(snapshot.disks, &:available_bytes)
      }
    end

    # A helper for snapshot. Array#sum is in activesupport, but not in the standard library.
    def sum(ary)
      ary.inject(0) { |sum, e| sum + yield(e) }
    end
  end
end
